local mod = {} -- will be filled with public functions

local settings = require('common.settings')
local colors = settings.colors
local fonts = settings.fonts
local players = require('common.players')

local players_movement_info = table.new(50, 0)

-- runs even if app is hidden, use it only if needed
function mod.update()
    local leaderboard = players.get_leaderboard()
    for i=0, #leaderboard-1 do
        local car_info = leaderboard[i].car
        local carpos = car_info.position
        local car_index = car_info.index
        local old_info = players_movement_info[car_index]
        if old_info == nil then old_info = table.new(0, 2) end
        
        local oldpos = old_info.old_carpos
        if oldpos ~= nil then
            local offset = (carpos - oldpos)
            local dist_diff = offset:length()
            if dist_diff > 0.5 then
                old_info.movement_change_time = Time
                old_info.old_carpos = carpos:clone()
            end
        else
            old_info.movement_change_time = Time
            old_info.old_carpos = carpos:clone()
        end
        
        players_movement_info[car_index] = table.clone(old_info, true)
    end
end

local map_pointcount = 560
local map_points = table.new(map_pointcount, 0)
local rotator = quat(0, 0, 0, 1)
local map_center = vec3()

local min_pos = vec2()
local max_pos = vec2()
function mod.init()
    local sim_info = ac.getSim()
    local track_length = sim_info.trackLengthM
    if track_length > 7000 then
        map_pointcount = math.max(math.floor(track_length / 40), map_pointcount)
    end
    
    
    local track_start_spline = players.get_sector_splits()[0]
    local worldpos1 = ac.trackCoordinateToWorld(vec3(0, 0, track_start_spline))
    local worldpos2 = ac.trackCoordinateToWorld(vec3(0, 0, track_start_spline + 0.01))
    local worldpos1_2d = vec2(worldpos1.x, worldpos1.z)
    local worldpos2_2d = vec2(worldpos2.x, worldpos2.z)
    local world_dir = (worldpos2_2d - worldpos1_2d):normalize()
    local angle = math.atan2(world_dir.y, world_dir.x) -- will be off, we'll fix it later
    
    local max_3d = vec3()
    local min_3d = vec3()
    for i=0, map_pointcount-1 do
        local world_point_pos = ac.trackCoordinateToWorld(vec3(0, 0, i / map_pointcount))
        
        if world_point_pos.x > max_3d.x then max_3d.x = world_point_pos.x end
        if world_point_pos.y > max_3d.y then max_3d.y = world_point_pos.y end
        if world_point_pos.z > max_3d.z then max_3d.z = world_point_pos.z end
        if world_point_pos.x < min_3d.x then min_3d.x = world_point_pos.x end
        if world_point_pos.y < min_3d.y then min_3d.y = world_point_pos.y end
        if world_point_pos.z < min_3d.z then min_3d.z = world_point_pos.z end
    end
    
    map_center = (max_3d + min_3d) / 2
    
    -- some tracks are run clockwise and others counterclockwise, the angle above
    -- will be off by 180 degrees in one of those 2 cases!
    -- to fix it we check in which case we are and add 180 degrees if needed
    -- solution source: answer by Sean the Bean: https://stackoverflow.com/questions/1165647/how-to-determine-if-a-list-of-polygon-points-are-in-clockwise-order
    local is_clockwise = false
    local the_sum = 0
    for i=0, map_pointcount-1 do
        local worldpos_curr = ac.trackCoordinateToWorld(vec3(0, 0, i / map_pointcount))
        local worldpos_next = ac.trackCoordinateToWorld(vec3(0, 0, ((i + 1) % map_pointcount) / map_pointcount))
        the_sum = the_sum + (worldpos_curr.x * worldpos_next.z - worldpos_next.x * worldpos_curr.z)
    end
    if the_sum > 0 then is_clockwise = true end
    if is_clockwise then
        rotator = quat():setAngleAxis(angle + math.pi, 0, 1, 0)
    else
        rotator = quat():setAngleAxis(angle, 0, 1, 0)
    end
    
    -- now that we can rotate correctly we can store map data pre-rotated (so we don't to these calculations every frame!)
    for i=0, map_pointcount-1 do
        local world_point_pos = ac.trackCoordinateToWorld(vec3(0, 0, i / map_pointcount))
        local offset = world_point_pos - map_center
        local rotated = offset:rotate(rotator)
        local new_pos = map_center + rotated
        
        -- recalculate bounds after rotation as if we never rotated anything
        if new_pos.x > max_pos.x then max_pos.x = new_pos.x end
        if new_pos.z > max_pos.y then max_pos.y = new_pos.z end
        if new_pos.x < min_pos.x then min_pos.x = new_pos.x end
        if new_pos.z < min_pos.y then min_pos.y = new_pos.z end
        map_points[i] = vec2(new_pos.x, new_pos.z)
    end
end

local function world_point_remap(to_remap)
    return vec2(
        settings.remap(to_remap.x, min_pos.x, max_pos.x, 0, 1),
        settings.remap(to_remap.y, min_pos.y, max_pos.y, 0, 1)
    )
end

local function world_point_3d_remap(to_remap)
    return vec2(
        settings.remap(to_remap.x, min_pos.x, max_pos.x, 0, 1),
        settings.remap(to_remap.z, min_pos.y, max_pos.y, 0, 1)
    )
end

local function center_to_map(normalized, map_center, mapsize, map_aspect)
    -- normalized goes from 0,0 to 1,1
    if map_aspect > 1 then
        local corner_tl = map_center - vec2(mapsize, mapsize / map_aspect) / 2
        return corner_tl + normalized * vec2(mapsize, mapsize / map_aspect)
    else
        local corner_tl = map_center - vec2(mapsize * map_aspect, mapsize) / 2
        return corner_tl + normalized * vec2(mapsize * map_aspect, mapsize)
    end
end

local function draw_map_path(center, mapsize, map_aspect, map_start, map_end)
    ui.pathClear()
    local idx_start = math.floor(map_start * map_pointcount)
    local idx_end = math.ceil(map_end * map_pointcount)
    for i= idx_start, idx_end-1 do
        local current_point = map_points[i]
        local normalized = world_point_remap(current_point)
        ui.pathLineTo(center_to_map(normalized, center, mapsize, map_aspect))
    end
end

local function rotate_mappos(mappos)
    local offset = mappos - map_center
    local rotated_offset = offset:rotate(rotator)
    return map_center + rotated_offset
end

local function draw_driver_dot(adjusted_pos, size, radius, lerp_t, car_racepos, fontsize, is_in_pitlane, wheels_outside, border_radius, dot_color, text_color, show_text_lappeds)
    if lerp_t <= 0 and not is_in_pitlane then
        ui.drawCircleFilled(adjusted_pos, radius, dot_color, 20)
    else
        ui.drawRectFilled(adjusted_pos - size, adjusted_pos + size, dot_color, radius)
    end
    if lerp_t <= 0 and show_text_lappeds then
        local number_size = ui.measureDWriteText(car_racepos, fontsize)
        ui.dwriteDrawText(car_racepos, fontsize, adjusted_pos - number_size / 2, text_color)
    end
    if is_in_pitlane == false and wheels_outside >= 4 then
        ui.drawCircle(adjusted_pos, radius + border_radius / 2, colors.TEXT_YELLOW, 12, border_radius)
    end
end

local on_show_animation_start = 0
local is_showing = true
local is_paused = false
function mod.on_open()
    if is_paused == false then
        on_show_animation_start = Time
        is_showing = true
    end
    is_paused = false
end

function mod.on_close()
    is_paused = ac.getSim().isPaused
    if is_paused == false then
        is_showing = false
    end
end

local function get_flag_color_and_width(flagtype)
    local map_color = colors.WHITE
    local width = 10
    if flagtype == nil then
        map_color = colors.WHITE
    elseif flagtype == ac.FlagType.Caution then
        map_color = colors.YELLOW
        width = 14
    elseif flagtype == ac.FlagType.Stop then
        map_color = colors.BLACK
        width = 14
    elseif flagtype == ac.FlagType.ReturnToPits then
        map_color = colors.RED
        width = 14
    elseif flagtype == ac.FlagType.FasterCar then
        map_color = colors.BLUE
        width = 14
    elseif flagtype == ac.FlagType.OneLapLeft then
        map_color = colors.WHITE
    elseif flagtype == ac.FlagType.Finished then
        map_color = colors.WHITE
    end
    return map_color, width * MapScale
end

local map_anim_start_time = Time
local old_flag_type = nil
local old_map_color = colors.WHITE
local old_width = 0
function mod.main()
    local draw_top_left = vec2(0, 22)
    local draw_size = ui.windowSize() - vec2(0, 22)
    local draw_center = draw_top_left + draw_size / 2
    
    local map_padding = 20 * MapScale
    local corner_tl = draw_center - draw_size / 2
    local mapsize = 270 * MapScale
    local dotsize = 300 * MapScale
    
    local map_full_width  = max_pos.x - min_pos.x
    local map_full_height = max_pos.y - min_pos.y
    local map_aspect_ratio = (map_full_width / map_full_height)
    
    local map_center = corner_tl + mapsize / 2 + vec2(map_padding, map_padding)
    settings.play_intro_anim_setup(map_center, vec2(dotsize, dotsize) * 1.4, on_show_animation_start, is_showing)
    ui.drawImage(
        settings.get_asset("map_bg"),
        map_center - dotsize / 2,
        map_center + dotsize / 2
    )
    local sim_info = ac.getSim()
    
    local flagtype = sim_info.raceFlagType
    local map_color, new_width = get_flag_color_and_width(flagtype)
    if old_flag_type == nil or old_flag_type ~= flagtype then
        old_map_color, old_width = get_flag_color_and_width(old_flag_type)
        old_flag_type = flagtype
        map_anim_start_time = Time
    end
    -- DEBUG(cogno): map area
    -- ui.drawRect(map_center - mapsize / 2, map_center + mapsize / 2, colors.BLUE)
    local anim_t = math.clamp(Time - map_anim_start_time, 0, 0.5) * 2
    local can_loop = false
    if anim_t >= 1 then can_loop = true end

    draw_map_path(map_center, mapsize, map_aspect_ratio, 0, anim_t)
    ui.pathStroke(map_color, can_loop, new_width * MapScale)
    if anim_t < 1 then
        draw_map_path(map_center, mapsize, map_aspect_ratio, anim_t, 1)
        ui.pathStroke(old_map_color, false, old_width * MapScale)
    end
    draw_map_path(map_center, mapsize, map_aspect_ratio, 0, 1)
    ui.pathStroke(rgbm(43 / 255, 43 / 255, 43 / 255, 1), true, 6 * MapScale)
    
    -- sectors lines
    local sectors = players.get_sector_splits()
    for i=0, #sectors-1 do
        local sector_percentage = sectors[i]
        
        -- PERF(cogno): I'm sure this can be made faster (I don't think left/right need to go
        -- through the whole chain, we just need the direction), so maybe this can get faster,
        -- but most tracks only have 3 sectors so I don't think it's that much gain...
        local road_right  = ac.trackCoordinateToWorld(vec3( 1, 0, sector_percentage))
        local road_left   = ac.trackCoordinateToWorld(vec3(-1, 0, sector_percentage))
        local road_center = ac.trackCoordinateToWorld(vec3(0, 0, sector_percentage))
        local line_left   = world_point_3d_remap(rotate_mappos(road_left))
        local line_right  = world_point_3d_remap(rotate_mappos(road_right))
        local line_center = world_point_3d_remap(rotate_mappos(road_center))
        local adjusted_left  = center_to_map(line_left,   map_center, mapsize, map_aspect_ratio)
        local adjusted_right = center_to_map(line_right,  map_center, mapsize, map_aspect_ratio)
        local adj_center     = center_to_map(line_center, map_center, mapsize, map_aspect_ratio)
        local adj_dir = (adjusted_right - adjusted_left):normalize()

        local sector_color = map_color
        local line_width = 3 * MapScale
        local line_len = 4 * MapScale
        if i==0 then
            sector_color = colors.WARNING_RED
            line_width = 5 * MapScale
            line_len = new_width * MapScale / 2
        end
        
        local p1 = adj_center + adj_dir * line_len
        local p2 = adj_center - adj_dir * line_len
        ui.drawLine(p1, p2, sector_color, line_width)
    end
    
    local leaderboard = players.get_leaderboard()
    local players_circle_radius = 11 * MapScale
    local players_circle_radius_min = 6 * MapScale
    local players_stopped_radius = 4 * MapScale
    local fontsize = settings.fontsize(9) * MapScale
    local big_fontsize = settings.fontsize(10) * MapScale
    ui.pushDWriteFont(fonts.archivo_medium)
    local my_car_info = ac.getCar(0)
    local player_lapcount = players.get_lapcount(0)
    local player_spline_total = player_lapcount + my_car_info.splinePosition
    local border_radius = 2
    
    local player_lb_index = players.get_player_leaderboard_index()
    local player_racepos = players.get_racepos(player_lb_index)
    
    local session = players.get_current_session()
    if session.type ~= ac.SessionType.Practice and session.type ~= ac.SessionType.Qualify then
        -- first draw lapped people
        for i=0, #leaderboard-1 do
            local car_info = leaderboard[i].car
            local car_worldpos = rotate_mappos(car_info.position)
            local player_normalized_trackpos = world_point_3d_remap(car_worldpos)
            local adjusted_pos = center_to_map(player_normalized_trackpos, map_center, mapsize, map_aspect_ratio)
            local car_racepos = players.get_racepos(i)
            
            if car_info.isConnected then
                local driver_lapcount = players.get_lapcount(car_info.index)
                local car_spline_total = driver_lapcount + car_info.splinePosition
                local spline_delta = player_spline_total - car_spline_total
                if (math.abs(spline_delta) >= 1) and car_info.index ~= 0 then
                    local dot_color = rgbm(0 / 255, 156 / 255, 255 / 255, 0)
                    if spline_delta <= -1 then
                        dot_color.mult = 1
                    else
                        dot_color.mult = 0.5
                    end
                    
                    local player_move_data = players_movement_info[car_info.index]
                    local elapsed = Time - player_move_data.movement_change_time
                    local lerp_t = math.clamp(settings.remap(elapsed, 3, 3.1, 0, 1), 0, 1)
                    local radius = math.lerp(players_circle_radius, players_stopped_radius, lerp_t)
                    local size = math.lerp(players_circle_radius, players_stopped_radius, lerp_t)
                    if car_info.isInPitlane and lerp_t <= 0 then
                        radius = players_circle_radius_min / 2
                        size = players_circle_radius_min
                    end
                    if Map_LowProfile == false then
                        draw_driver_dot(adjusted_pos, size, radius, lerp_t, car_racepos, fontsize, car_info.isInPitlane, car_info.wheelsOutside, border_radius, dot_color, colors.WHITE, spline_delta <= -1)
                    else
                        local pos_dist = math.abs(car_racepos - player_racepos)
                        if pos_dist <= 2 then
                            draw_driver_dot(adjusted_pos, size, radius, lerp_t, car_racepos, fontsize, car_info.isInPitlane, car_info.wheelsOutside, border_radius, dot_color, colors.WHITE, spline_delta <= -1)
                        else
                            ui.drawRectFilled(adjusted_pos - players_stopped_radius, adjusted_pos + players_stopped_radius, dot_color, radius)
                        end
                    end
                end
            end
        end
    end
    
    -- then normal people!
    for i=0, #leaderboard-1 do
        local car_info = leaderboard[i].car
        local car_worldpos = rotate_mappos(car_info.position)
        local player_normalized_trackpos = world_point_3d_remap(car_worldpos)
        local adjusted_pos = center_to_map(player_normalized_trackpos, map_center, mapsize, map_aspect_ratio)
        local car_racepos = players.get_racepos(i)
        local car_spline_total = players.get_lapcount(car_info.index) + car_info.splinePosition
        local spline_delta = player_spline_total - car_spline_total
        
        if car_info.isConnected == true then
            local to_draw = (math.abs(spline_delta) <= 1)  and car_info.index ~= 0 and car_racepos ~= 1
            if session.type == ac.SessionType.Practice or session.type == ac.SessionType.Qualify then
                to_draw = true
            end
            local player_move_data = players_movement_info[car_info.index]
            if to_draw and player_move_data ~= nil then
                local elapsed = Time - player_move_data.movement_change_time
                local lerp_t = math.clamp(settings.remap(elapsed, 3, 3.1, 0, 1), 0, 1)
                local radius = math.lerp(players_circle_radius, players_stopped_radius, lerp_t)
                local size = math.lerp(players_circle_radius, players_stopped_radius, lerp_t)
                if car_info.isInPitlane and lerp_t <= 0 then
                    radius = players_circle_radius_min / 2
                    size = players_circle_radius_min
                end
                if Map_LowProfile == false then
                    draw_driver_dot(adjusted_pos, size, radius, lerp_t, car_racepos, fontsize, car_info.isInPitlane, car_info.wheelsOutside, border_radius, colors.LIGHT_GREEN, colors.DARK_BG, true)
                else
                    local pos_dist = math.abs(car_racepos - player_racepos)
                    if pos_dist <= 2 then
                        draw_driver_dot(adjusted_pos, size, radius, lerp_t, car_racepos, fontsize, car_info.isInPitlane, car_info.wheelsOutside, border_radius, colors.LIGHT_GREEN, colors.DARK_BG, true)
                    else
                        ui.drawRectFilled(adjusted_pos - players_stopped_radius, adjusted_pos + players_stopped_radius, colors.LIGHT_GREEN, radius)
                    end
                end
            end
        end
    end
    
    -- then the first one of the race
    ui.pushDWriteFont(fonts.archivo_bold) -- both first person and player are important
    for i=0, #leaderboard-1 do
        local car_info = leaderboard[i].car
        local car_worldpos = rotate_mappos(car_info.position)
        local player_normalized_trackpos = world_point_3d_remap(car_worldpos)
        local adjusted_pos = center_to_map(player_normalized_trackpos, map_center, mapsize, map_aspect_ratio)
        local car_racepos = players.get_racepos(i)
        
        if car_info.isConnected == true then
            local player_move_data = players_movement_info[car_info.index]
            if i == 0 and player_move_data ~= nil then
                local elapsed = Time - player_move_data.movement_change_time
                local lerp_t = math.clamp(settings.remap(elapsed, 3, 3.1, 0, 1), 0, 1)
                local radius = math.lerp(players_circle_radius, players_stopped_radius, lerp_t)
                local size = math.lerp(players_circle_radius, players_stopped_radius, lerp_t)
                if car_info.isInPitlane and lerp_t <= 0 then
                    radius = players_circle_radius_min / 2
                    size = players_circle_radius_min
                end
                
                -- even when the map is minimized we still need to show him, so we skip those checks
                draw_driver_dot(adjusted_pos, size, radius, lerp_t, car_racepos, fontsize, car_info.isInPitlane, car_info.wheelsOutside, border_radius, colors.PURPLE, colors.WHITE, true)
            end
        end
    end
    
    -- then the player!
    local players_main_size = 25 * MapScale
    for i=0, #leaderboard-1 do
        local car_info = leaderboard[i].car
        local car_worldpos = rotate_mappos(car_info.position)
        local player_normalized_trackpos = world_point_3d_remap(car_worldpos)
        local adjusted_pos = center_to_map(player_normalized_trackpos, map_center, mapsize, map_aspect_ratio)
        local car_racepos = players.get_racepos(i)
        
        if car_info.index == 0 then
            local color = colors.TEXT_YELLOW
            ui.drawCircleFilled(adjusted_pos, players_main_size / 2, color, 20)
            local number_size = ui.measureDWriteText(car_racepos, big_fontsize)
            ui.dwriteDrawText(car_racepos, big_fontsize, adjusted_pos - number_size / 2, colors.DARK_BG)
            -- NOTE(cogno): no need to draw the yellow circle for yourself, we don't care!
        end
    end
    ui.popDWriteFont()
    ui.popDWriteFont()
    
    settings.play_intro_anim(map_center, vec2(dotsize, dotsize) * 1.4, on_show_animation_start, MapScale)
    settings.lock_app(map_center, vec2(dotsize, dotsize) * 1.4, "CMRT map", MapScale)
    settings.auto_scale_window(vec2(dotsize, dotsize) * 1.02, "CMRT map")
    settings.auto_place_once(vec2(dotsize, dotsize), "CMRT map")
    --ui.drawRect(draw_center + vec2(-1, -1) * mapsize, draw_center + vec2(map_aspect_ratio, 1) * mapsize, colors.BLUE)
    
    -- DEBUG(cogno): draw area
    -- ui.drawRect(draw_top_left, draw_top_left + draw_size, colors.WHITE)
end

return mod -- expose functions to the outside